#include <err.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netgraph.h>
#include <netgraph/ng_message.h>
#include <netgraph/ng_socket.h>
#include <netgraph/ng_tee.h>
#include <netgraph/ng_tty.h>

#include "../ng_irlap_async/ng_irlap_async.h"
#include "../ng_irlap/ng_irlap.h"

#define TTYDEV "/dev/cuaa1"
#define NGDEBUG 3

void
check_module(const char *mod_name)
{
	if (modfind(mod_name) < 0) {
		if (kldload(mod_name) < 0)
			err(1, "%s: Kernel module not available", mod_name);
	}
}
	
void
usage(const char *name)
{
	fprintf(stderr, "Usage: %s [-d]\n", name);
	exit(1);
}

int
main(int argc, char **argv)
{
	int fd, ret, ldisc = NETGRAPHDISC;
	int hotchar = 0xc1;
	int csp, dsp;
	struct nodeinfo ni;
	struct ngm_mkpeer mkp;
	struct ngm_connect ngc;
	struct ng_irlap_async_cfg nac;
	struct ng_mesg nm;
	struct ng_irlap_device_info di;
	char path[NG_PATHLEN + 1], hostname[MAXHOSTNAMELEN + 1], *c;
	u_char buf[1024];

	if (gethostname(hostname, MAXHOSTNAMELEN + 1) < 0)
		err(1, "gethostname");

	c = strchr(hostname, '.');
	*c = '\0';

	check_module("netgraph");
	check_module("ng_socket");
	check_module("ng_tee");
	check_module("ng_tty");
	check_module("ng_irlap");
	check_module("ng_irlap_async");

	NgSetDebug(NGDEBUG);

	/* Create our ng_socket endpoint node. */
	ret = NgMkSockNode("irdad", &csp, &dsp);
	if (ret < 0) {
		perror("NgMkSockNode");
		exit(1);
	}

	/* Create our ng_tty node. */
	fd = open(TTYDEV, O_RDWR);
	ret = ioctl(fd, TIOCSETD, &ldisc);
	if (ret < 0) {
		perror("set line discipline");
		exit(1);
	}
	ret = ioctl(fd, NGIOCGINFO, &ni);
	if (ret < 0) {
		perror("query node info");
		exit(1);
	}

	/* Configure our ng_tty node. */
	sprintf(path, "%s:", ni.name);
	ret = NgSendMsg(csp, path, NGM_TTY_COOKIE, NGM_TTY_SET_HOTCHAR,
	    &hotchar, sizeof(hotchar));
	if (ret < 0) {
		perror("set ng_tty hot char");
		exit(1);
	}

	/* Create an ng_irlap_async node connected to our ng_tty node. */
	snprintf(mkp.type, sizeof(mkp.type), "%s", NG_IRLAP_ASYNC_NODE_TYPE);
	snprintf(mkp.ourhook, sizeof(mkp.ourhook), "hook");
	snprintf(mkp.peerhook, sizeof(mkp.peerhook), "%s",
		NG_IRLAP_ASYNC_HOOK_ASYNC);
	ret = NgSendMsg(csp, path, NGM_GENERIC_COOKIE, NGM_MKPEER, &mkp,
	    sizeof(mkp));
	if (ret < 0) {
		perror("create irlap_async node");
		exit(1);
	}

	/*
	 * Create an ng_tee node, and connect one end to the irlap_async node
	 * and the monitor channels to us.
	 */
	sprintf(path, "%s:hook", ni.name);
	snprintf(mkp.type, sizeof(mkp.type), "%s", NG_TEE_NODE_TYPE);
	snprintf(mkp.ourhook, sizeof(mkp.ourhook), NG_IRLAP_ASYNC_HOOK_SYNC);
	snprintf(mkp.peerhook, sizeof(mkp.peerhook), "%s", NG_TEE_HOOK_RIGHT);
	ret = NgSendMsg(csp, path, NGM_GENERIC_COOKIE, NGM_MKPEER, &mkp,
	    sizeof(mkp));
	if (ret < 0) {
		perror("create tee node");
		exit(1);
	}

	snprintf(ngc.path, sizeof(ngc.path), "%s.sync", path);
	snprintf(ngc.ourhook, sizeof(ngc.ourhook), "dev");
	snprintf(ngc.peerhook, sizeof(ngc.peerhook), NG_TEE_HOOK_RIGHT2LEFT);
	ret = NgSendMsg(csp, ".:", NGM_GENERIC_COOKIE, NGM_CONNECT, &ngc,
	    sizeof(ngc));
	if (ret < 0) {
		perror("connect dev to right2left");
		exit(1);
	}

	snprintf(ngc.ourhook, sizeof(ngc.ourhook), "irlap");
	snprintf(ngc.peerhook, sizeof(ngc.peerhook), NG_TEE_HOOK_LEFT2RIGHT);
	ret = NgSendMsg(csp, ".:", NGM_GENERIC_COOKIE, NGM_CONNECT, &ngc,
	    sizeof(ngc));
	if (ret < 0) {
		perror("connect irlap to left2right");
		exit(1);
	}

	/* Fire up the ng_irlap node */
	sprintf(path, "%s:hook.sync", ni.name);
	snprintf(mkp.type, sizeof(mkp.type), "%s", NG_IRLAP_NODE_TYPE);
	snprintf(mkp.ourhook, sizeof(mkp.ourhook), NG_TEE_HOOK_LEFT);
	snprintf(mkp.peerhook, sizeof(mkp.peerhook), "%s", NG_IRLAP_HOOK_DOWN);
	ret = NgSendMsg(csp, path, NGM_GENERIC_COOKIE, NGM_MKPEER, &mkp,
	    sizeof(mkp));
	if (ret < 0) {
		perror("create irlap node");
		exit(1);
	}

	/* Attempt to set our device info */
	bzero(&di, sizeof(di));
	di.data[0] = 0x82;
	di.data[1] = 0x01;
	di.data[2] = 0x00;
	bcopy(hostname, &di.data[3], strlen(hostname));
	di.length = strlen(hostname) + 3;

	sprintf(path, "%s:hook.sync.left", ni.name);
	ret = NgSendMsg(csp, path, NGM_IRLAP_COOKIE,
	    NGM_IRLAP_CMD_SET_DEVICE_INFO, &di, 28);
	if (ret < 0)
		err(1, "set device info");

	/* Start processing data. */
	for (;;) {
		char hook[NG_HOOKLEN + 1];

		ret = NgRecvData(dsp, buf, 1024, hook);
		if (ret < 0) {
			perror("NgRecvData");
			exit(1);
		}
	};

	close(fd);
}
