#include <err.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netgraph.h>
#include <netgraph/ng_message.h>
#include <netgraph/ng_socket.h>
#include <netgraph/ng_tty.h>

#include "irlapasync/ng_irlapasync.h"
#include "irlap.h"

#define TTYDEV "/dev/cuaa1"
#define NGDEBUG 3

static u_int8_t my_info[] = { 0x82, 0x20, 0x00, 0x72, 0x61, 0x74, 0x2d, 0x74,
	0x68, 0x69, 0x6e, 0x67 };
int my_info_len = 12;

static struct irlap_state state;

static int process_frame(struct irlap_frame *, int);
static int process_xid(struct irlap_frame *, int);
static int process_test(struct irlap_frame *, int);
static int process_snrm(struct irlap_frame *, int);
static int process_disc(struct irlap_frame *, int);
static int process_supervisory(struct irlap_frame *, int);

void
check_module(const char *mod_name)
{
	if (modfind(mod_name) < 0) {
		if (kldload(mod_name) < 0)
			err(1, "%s: Kernel module not available", mod_name);
	}
}
	
void
usage(const char *name)
{
	fprintf(stderr, "Usage: %s [-d]\n", name);
	exit(1);
}

int
main(int argc, char **argv)
{
	int fd, ret, ldisc = NETGRAPHDISC;
	int hotchar = 0xc1;
	int discover = 0;
	struct nodeinfo ni;
	struct ngm_mkpeer mkp;
	struct ngm_connect ngc;
	struct ng_irlapasync_cfg nac;
	struct ng_mesg nm;
	char path[NG_PATHLEN + 1];
	u_char buf[1024];

	extern char *optarg;
	extern int optind;
	int ch;

	while ((ch = getopt(argc, argv, "d")) != -1) {
		switch (ch) {
		case 'd':
			discover = 1;
			break;
		default:
			usage(argv[0]);
		}
	}
	argc -= optind;
	argv += optind;

	check_module("netgraph");
	check_module("ng_socket");
	check_module("ng_tty");
	check_module("ng_irlapasync");

	srandom();
	state.is_address = (u_int32_t) random();
	state.is_state = IRLAP_STATE_NDM;
	state.is_conn_params.rate = 0x02;
	state.is_conn_params.maxturn = 0x01;
	state.is_conn_params.datasize = 0x0f;
	state.is_conn_params.winsize = 0x01;
	state.is_conn_params.nbof = 0x08;
	state.is_conn_params.minturn = 0x07;
	state.is_conn_params.thresh = 0xff;

	NgSetDebug(NGDEBUG);

	/* Create our ng_socket endpoint node. */
	ret = NgMkSockNode("irdad", &state.is_csp, &state.is_dsp);
	if (ret < 0) {
		perror("NgMkSockNode");
		exit(1);
	}

	/* Create our ng_tty node. */
	fd = open(TTYDEV, O_RDWR);
	ret = ioctl(fd, TIOCSETD, &ldisc);
	if (ret < 0) {
		perror("set line discipline");
		exit(1);
	}
	ret = ioctl(fd, NGIOCGINFO, &ni);
	if (ret < 0) {
		perror("query node info");
		exit(1);
	}

	/* Configure our ng_tty node. */
	sprintf(path, "%s:", ni.name);
	ret = NgSendMsg(state.is_csp, path, NGM_TTY_COOKIE, NGM_TTY_SET_HOTCHAR,
		&hotchar, sizeof(hotchar));
	if (ret < 0) {
		perror("set ng_tty hot char");
		exit(1);
	}

	/* Create an ng_irlapasync node connected to our ng_tty node. */
	snprintf(mkp.type, sizeof(mkp.type), "%s", NG_IRLAPASYNC_NODE_TYPE);
	snprintf(mkp.ourhook, sizeof(mkp.ourhook), "hook");
	snprintf(mkp.peerhook, sizeof(mkp.peerhook), "%s",
		NG_IRLAPASYNC_HOOK_ASYNC);
	ret = NgSendMsg(state.is_csp, path, NGM_GENERIC_COOKIE, NGM_MKPEER,
		&mkp, sizeof(mkp));
	if (ret < 0) {
		perror("create irlapasync node");
		exit(1);
	}

	/* Connect our ng_socket node to our ng_irlapasync node. */
	snprintf(ngc.path, sizeof(ngc.path), "%shook", path);
	snprintf(ngc.ourhook, sizeof(ngc.ourhook), "irlap");
	snprintf(ngc.peerhook, sizeof(ngc.peerhook), "sync");
	ret = NgSendMsg(state.is_csp, ".:", NGM_GENERIC_COOKIE, NGM_CONNECT,
		&ngc, sizeof(ngc));
	if (ret < 0) {
		perror("connect netgraph node");
		exit(1);
	}

	/* Start processing data. */
	do {
		char hook[NG_HOOKLEN + 1];
		int ret, i;
		struct pollfd pfds[1];

		do {
			pfds[0].fd = state.is_dsp;
			pfds[0].events = POLLRDNORM;

			ret = poll(pfds, 1, 80);
		} while (!(ret && (pfds[0].revents | POLLRDNORM)));

		ret = NgRecvData(state.is_dsp, buf, 1024, hook);
		if (ret < 0) {
			perror("NgRecvData");
			exit(1);
		}
	} while (process_frame((struct irlap_frame *) buf, ret));

	close(fd);
}

static int
process_frame(struct irlap_frame *frame, int framesize)
{
	switch (frame->ifr_control & ~IRLAP_MASK_PF) {
	case IRLAP_COMMAND_SNRM:	/* Also covers IRLAP_RESPONSE_SNRM */
		return(process_snrm(frame, framesize));
		break;
	case IRLAP_COMMAND_DISC:
		return(process_disc(frame, framesize));
		break;
	case IRLAP_COMMAND_UI:		/* Also covers IRLAP_RESPONSE_UI */
		printf("Received unimplemented UI command.\n");
		return(0);
		break;
	case IRLAP_COMMAND_XID:
		return(process_xid(frame, framesize));
		break;
	case IRLAP_COMMAND_TEST:	/* Also covers IRLAP_RESPONSE_TEST */
		return(process_test(frame, framesize));
		break;
	case IRLAP_RESPONSE_UA:
		printf("Received unimplemented UA response.\n");
		return(0);
		break;
	case IRLAP_RESPONSE_FRMR:
		printf("Received unimplemented FRMR response.\n");
		return(0);
		break;
	case IRLAP_RESPONSE_DM:
		printf("Received unimplemented DM response.\n");
		return(0);
		break;
	case IRLAP_RESPONSE_XID:
		printf("Received unimplemented XID response.\n");
		return(0);
		break;
	default:
		if (frame->ifr_control & 0x01 == 0) {
			printf("Implement I frames you dolt!\n");
			return(0);
		}
		else
			return(process_supervisory(frame, framesize));
	}
}

static int
process_xid(struct irlap_frame *frame, int framesize)
{
	struct xid_frame *xid;

	switch (state.is_state) {
	case IRLAP_STATE_NDM:
		if (*frame->ifr_data != 0x1) {
			printf("Unknown XID format: %02x\n", *frame->ifr_data);
			return(0);
		}

		xid = (struct xid_frame *) (frame->ifr_data + 1);

		switch (xid->xid_flags) {
		case 0x3:
			state.is_xid_slot = (u_int8_t) random() % 16;
			break;
		case 0x2:
			state.is_xid_slot = (u_int8_t) random() % 8;
			break;
		case 0x1:
			state.is_xid_slot = (u_int8_t) random() % 6;
			break;
		case 0x0:
		default:
			state.is_xid_slot = (u_int8_t) random() % 1;
			break;
		}

		state.is_state = IRLAP_STATE_REPLY;
		printf("NDM -> REPLY\n");

		if (state.is_xid_slot != 0)
			break;
	case IRLAP_STATE_REPLY:
		if (*frame->ifr_data != 0x1) {
			printf("Unknown XID format: %02x\n", *frame->ifr_data);
			return(0);
		}

		xid = (struct xid_frame *) (frame->ifr_data + 1);

		if (state.is_xid_slot >= xid->xid_slot) {
			int size = sizeof(struct irlap_frame) +
				sizeof(struct xid_frame) + 10;
			int i;
			u_char buf[size];
			struct irlap_frame *outframe;
			struct xid_frame *outxid;

			outframe = (struct irlap_frame *) buf;
			outxid = (struct xid_frame *) (outframe->ifr_data + 1);

			outframe->ifr_connection = IRLAP_BROADCAST;
			outframe->ifr_control =
				IRLAP_RESPONSE_XID | IRLAP_MASK_PF;
			outframe->ifr_data[0] = 1;	/* XID format */

			outxid->xid_source = state.is_address;
			outxid->xid_destination = xid->xid_source;
			outxid->xid_flags = 0;
			outxid->xid_slot = 0;
			outxid->xid_version = 0;

			memcpy(outxid->xid_data, my_info, my_info_len);

			NgSendData(state.is_dsp, "irlap", buf, size);

			state.is_xid_slot = 0;
		}

		if (xid->xid_slot == 0xff) {
			state.is_state = IRLAP_STATE_NDM;
			printf("REPLY -> NDM\n");
		}

		break;
	default:
		printf("Unknown state for XID frame: %d\n", state.is_state);
		break;
	}

	return(1);
}

static int
process_test(struct irlap_frame *frame, int framesize)
{
	u_char buf[26];
	struct irlap_frame *outframe;
	struct test_frame *test, *outtest;

	switch (state.is_state) {
	case IRLAP_STATE_NDM:
		test = (struct test_frame *) frame->ifr_data;
		if (test->test_destination != state.is_address)
			break;

		memcpy(buf, frame, 26);

		outframe = (struct irlap_frame *) buf;
		outframe->ifr_connection = IRLAP_BROADCAST;
		outframe->ifr_control = IRLAP_RESPONSE_TEST;

		outtest = (struct test_frame *) outframe->ifr_data;
		outtest->test_destination = outtest->test_source;
		outtest->test_source = state.is_address;

		NgSendData(state.is_dsp, "irlap", buf, 26);

		break;
	default:
		break;
	}

	return 1;
}

static int
process_snrm(struct irlap_frame *frame, int framesize)
{
	struct irlap_frame *outframe;
	struct snrm_frame *snrm;
	struct ua_frame *ua;
	u_char buf[31];
	int i = 0;

	snrm = (struct snrm_frame *) frame->ifr_data;
	if (snrm->snrm_destination != state.is_address)
		return(0);

	switch (state.is_state) {
	case IRLAP_STATE_NDM:
		state.is_connection = snrm->snrm_connection & ~IRLAP_MASK_CR;
		printf(">>> %x\n", state.is_connection);

		outframe = (struct irlap_frame *) buf;
		outframe->ifr_connection = snrm->snrm_connection;
		outframe->ifr_control = IRLAP_RESPONSE_UA | IRLAP_MASK_PF;

		/* XXX: Largely irrelevant atm. */
		state.is_state = IRLAP_STATE_CONN;
		printf("NDM -> CONN\n");

		ua = (struct ua_frame *) outframe->ifr_data;
		ua->ua_source = state.is_address;
		ua->ua_destination = snrm->snrm_source;

		ua->ua_data[i++] = IRLAP_PARAM_RATE;
		ua->ua_data[i++] = sizeof(state.is_conn_params.rate);
		ua->ua_data[i++] = state.is_conn_params.rate;

		ua->ua_data[i++] = IRLAP_PARAM_MAXTURN;
		ua->ua_data[i++] = sizeof(state.is_conn_params.maxturn);
		ua->ua_data[i++] = state.is_conn_params.maxturn;

		ua->ua_data[i++] = IRLAP_PARAM_DATASIZE;
		ua->ua_data[i++] = sizeof(state.is_conn_params.datasize);
		ua->ua_data[i++] = state.is_conn_params.datasize;

		ua->ua_data[i++] = IRLAP_PARAM_WINSIZE;
		ua->ua_data[i++] = sizeof(state.is_conn_params.winsize);
		ua->ua_data[i++] = state.is_conn_params.winsize;

		ua->ua_data[i++] = IRLAP_PARAM_NBOF;
		ua->ua_data[i++] = sizeof(state.is_conn_params.nbof);
		ua->ua_data[i++] = state.is_conn_params.nbof;

		ua->ua_data[i++] = IRLAP_PARAM_MINTURN;
		ua->ua_data[i++] = sizeof(state.is_conn_params.minturn);
		ua->ua_data[i++] = state.is_conn_params.minturn;

		ua->ua_data[i++] = IRLAP_PARAM_THRESH;
		ua->ua_data[i++] = sizeof(state.is_conn_params.thresh);
		ua->ua_data[i++] = state.is_conn_params.thresh;

		state.is_state = IRLAP_STATE_NRM_S;
		printf("CONN -> NRM(S)\n");
		state.is_remote = snrm->snrm_source;

		NgSendData(state.is_dsp, "irlap", buf, 31);
	}

	return (1);
}

static int
process_supervisory(struct irlap_frame *frame, int framesize)
{
	unsigned char buf[sizeof(struct irlap_frame)];
	struct irlap_frame *outframe = (struct irlap_frame *) buf;

	outframe->ifr_connection = state.is_connection;
	outframe->ifr_control = 0x11;

	NgSendData(state.is_dsp, "irlap", buf, sizeof(struct irlap_frame));

	return(1);
}

static int
process_disc(struct irlap_frame *frame, int framesize)
{
	unsigned char buf[sizeof(struct irlap_frame)];
	struct irlap_frame *outframe = (struct irlap_frame *) buf;

	if (state.is_state == IRLAP_STATE_NRM_S) {
		outframe->ifr_connection = frame->ifr_connection;
		outframe->ifr_control = IRLAP_RESPONSE_UA;
	
		NgSendData(state.is_dsp, "irlap", buf,
			sizeof(struct irlap_frame));

		state.is_state = IRLAP_STATE_NDM;
		printf("NRM(S) -> NDM\n");
		state.is_connection = 0;
		state.is_remote = 0;
	} else
		printf("Unknown state for DISC frame: %d\n", state.is_state);
	return(1);
}
