/* $Id: ng_irlap.c,v 1.2 2000/08/03 13:12:47 benno Exp $ */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/file.h>
#include <sys/errno.h>

#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_parse.h>

#include "../ng_irlap_async/ng_irlap_async.h"
#include "ng_irlap.h"

/* State values */
enum {
	NG_IRLAP_STATE_NDM = 1,
	NG_IRLAP_STATE_QUERY,
	NG_IRLAP_STATE_REPLY,
	NG_IRLAP_STATE_CONN,
	NG_IRLAP_STATE_SETUP,
	NG_IRLAP_STATE_NRM,
};

/* NRM state values */
enum {
	NG_IRLAP_NRM_XMIT = 1,
	NG_IRLAP_NRM_RECV,
	NG_IRLAP_NRM_ERROR,
	NG_IRLAP_NRM_RESET_CHECK,
	NG_IRLAP_NRM_RESET,
	NG_IRLAP_NRM_BUSY,
	NG_IRLAP_NRM_BUSY_WAIT,
	NG_IRLAP_NRM_SCLOSE,
};

/* Masks */
#define	NG_IRLAP_MASK_CR	0x01	/* Command/Response bit */
#define	NG_IRLAP_MASK_PF	0x10	/* Poll/Final bit */
#define NG_IRLAP_MASK_NR	0xe0	/* Receive sequence number (Nr) */
#define NG_IRLAP_MASK_NS	0x0e	/* Send sequence number (Ns) */
#define NG_IRLAP_MASK_SCMD	0x0c	/* Supervisory command */

/* Neat macros to get PF/NR/NS/SCMD values out of a control field */
#define NG_IRLAP_GET_PF(control)	(control & NG_IRLAP_MASK_PF) >> 4
#define NG_IRLAP_GET_NR(control)	(control & NG_IRLAP_MASK_NR) >> 5
#define NG_IRLAP_GET_NS(control)	(control & NG_IRLAP_MASK_NS) >> 1
#define NG_IRLAP_GET_SCMD(control)	(control & NG_IRLAP_MASK_SCMD) >> 2

/* Timeout values, some defined in IrLAP p. 92 */
#define NG_IRLAP_TIMEOUT_POLL		((450 * hz) / 1000)
#define NG_IRLAP_TIMEOUT_FINAL		((500 * hz) / 1000)
#define NG_IRLAP_TIMEOUT_WATCHDOG	(20 * hz)
#define NG_IRLAP_TIMEOUT_CONN_WD	(NG_IRLAP_TIMEOUT_POLL * 2)
#define NG_IRLAP_TIMEOUT_MEDIABUSY	((500 * hz) / 1000)
#define NG_IRLAP_TIMEOUT_SLOT		((80 * hz) / 1000)
#define NG_IRLAP_TIMEOUT_QUERY		hz

/* Broadcast connection address */
#define	NG_IRLAP_BROADCAST	0xfe

/* U-frame commands */
#define	NG_IRLAP_COMMAND_SNRM	0x83
#define NG_IRLAP_COMMAND_DISC	0x43
#define	NG_IRLAP_COMMAND_UI	0x03
#define	NG_IRLAP_COMMAND_XID	0x2f
#define	NG_IRLAP_COMMAND_TEST	0xe3

/* U-frame responses */
#define	NG_IRLAP_RESPONSE_RNRM	0x83
#define	NG_IRLAP_RESPONSE_UA	0x63
#define	NG_IRLAP_RESPONSE_FRMR	0x87
#define	NG_IRLAP_RESPONSE_DM	0x0f
#define	NG_IRLAP_RESPONSE_UI	0x03
#define NG_IRLAP_RESPONSE_XID	0xaf
#define NG_IRLAP_RESPONSE_TEXT	0xe3

/* S-frame flags */
#define	NG_IRLAP_S_RR		0x00
#define NG_IRLAP_S_RNR		0x01
#define	NG_IRLAP_S_REJ		0x02
#define	NG_IRLAP_S_SREJ		0x03

/* Parameter types */
#define	NG_IRLAP_PARAM_RATE	0x01
#define NG_IRLAP_PARAM_MAXTURN	0x82
#define NG_IRLAP_PARAM_DATASIZE	0x83
#define	NG_IRLAP_PARAM_WINSIZE	0x84
#define NG_IRLAP_PARAM_NBOF	0x85
#define NG_IRLAP_PARAM_MINTURN	0x86
#define NG_IRLAP_PARAM_THRESH	0x08

/* An IrLAP frame */
struct ng_irlap_frame {
	u_int8_t	connection;
	u_int8_t	control;
	u_int8_t	data[0];
};

/* Format 1 XID frame */
struct ng_irlap_xid {
	u_int32_t	source;
	u_int32_t	destination;
	u_int8_t	flags;
	u_int8_t	slot;
	u_int8_t	version;
	u_int8_t	data[0];
};

/* TEST frame */
struct ng_irlap_test {
	u_int32_t	source;
	u_int32_t	destination;
	u_int8_t	data[0];
};

/* SNRM frame */
struct ng_irlap_snrm {
	u_int32_t	source;
	u_int32_t	destination;
	u_int8_t	connection;
	u_int8_t	data[0];
};

/* UA frame */
struct ng_irlap_ua {
	u_int32_t	source;
	u_int32_t	destination;
	u_int8_t	data[21];
};

/* FRMR frame */
struct ng_irlap_frmr {
	u_int8_t	control;
	u_int8_t	data[2];
};

/* Private data structure */
struct ng_irlap_private {
	node_p				node;		/* Our node */
	hook_p				up;		/* Upper layer */
	hook_p				down;		/* Lower layer */
	u_int8_t			state;		/* State */
	u_int8_t			nrm_state;	/* NRM State */
	u_int8_t			primary;	/* Are we primary or
								secondary? */
	u_int32_t			address;	/* IrLAP device
								address */
	u_int8_t			xid_slot;	/* Current XID slot */
	u_int8_t			connection;	/* Connection ID */
	struct ng_irlap_device_info	device_info;
	struct ng_irlap_conn_params	conn_params;

	/* Timeouts */
	struct callout_handle		watchdog;	/* Watchdog Timer */
	int				watchdog_expires;

	/* Sequence numbers */
	u_int8_t			in_seq;		/* Inbound sequence */
	u_int8_t			out_seq;	/* Outbound sequence */

	/* Flags */
	u_int8_t			media_busy;
	u_int8_t			remote_busy;
};
typedef struct ng_irlap_private *sc_p;

/* Handy macro */
#define ERROUT(x)               do { error = (x); goto done; } while (0)

/* Netgraph methods */
static ng_constructor_t		ngi_constructor;
static ng_rcvdata_t		ngi_rcvdata;
static ng_rcvmsg_t		ngi_rcvmsg;
static ng_shutdown_t		ngi_shutdown;
static ng_newhook_t		ngi_newhook;
static ng_disconnect_t		ngi_disconnect;

/* Data receive functions */
/*static int	ngi_rcv_up(const sc_p sc, struct mbuf *m, meta_p meta);*/
static int	ngi_rcv_down(const sc_p sc, struct mbuf *m, meta_p meta);

/* Data processing functions */
static int	ngi_process_info(const sc_p sc, struct mbuf *m, meta_p meta);
static int	ngi_process_supervisory(const sc_p sc, struct mbuf *m,
		    meta_p meta);
static int	ngi_process_unnumbered(const sc_p sc, struct mbuf *m,
		    meta_p meta);
static int	ngi_process_xid(const sc_p sc, struct mbuf *m, meta_p meta);
static int	ngi_process_snrm(const sc_p sc, struct mbuf *m, meta_p meta);
static int	ngi_process_disc(const sc_p sc, struct mbuf *m, meta_p meta);

/* Data send functions */
static int	ngi_send_supervisory(const sc_p sc, u_int8_t command,
		    u_int8_t cr, u_int8_t vr, u_int8_t pf);

/* Miscellaneous functions */
static void		ngi_start_watchdog(const sc_p sc);
static void		ngi_stop_watchdog(const sc_p sc);
static timeout_t	ngi_watchdog_expired;

/* Parse type for struct ng_irlap_conn_params */
static const struct ng_parse_struct_info
	ngi_conn_params_type_info = NG_IRLAP_CONN_PARAMS_TYPE_INFO;
static const struct ng_parse_type ngi_conn_params_type = {
	&ng_parse_struct_type,
	&ngi_conn_params_type_info
};

/* Get length of the data array inside a discovery entry struct */
static int
ng_irlap_device_info_data_getLength(const struct ng_parse_type *type,
	const u_char *start, const u_char *buf)
{
	const struct ng_irlap_device_info *di =
	    (const struct ng_irlap_device_info *)start;

	return di->length;
}

/* Parse type for a device info data array */
static const struct ng_parse_type ng_irlap_device_info_data_type = {
	&ng_parse_bytearray_type,
	&ng_irlap_device_info_data_getLength
};

/* Parse type for struct ng_irlap_device_info */
static const struct ng_parse_struct_info
	ngi_device_info_type_info = NG_IRLAP_DEVICE_INFO_TYPE_INFO;
static const struct ng_parse_type ngi_device_info_type = {
	&ng_parse_struct_type,
	&ngi_device_info_type_info
};

/* Parse type for struct ng_irlap_discovery_entry */
static const struct ng_parse_struct_info
	ngi_discovery_entry_type_info = NG_IRLAP_DISCOVERY_ENTRY_TYPE_INFO;
static const struct ng_parse_type ngi_discovery_entry_type = {
	&ng_parse_struct_type,
	&ngi_discovery_entry_type_info
};

/* List of commands and how to convert arguments to/from ASCII */
static const struct ng_cmdlist ngi_cmdlist[] = {
	{
	  NGM_IRLAP_COOKIE,
	  NGM_IRLAP_CMD_SET_DEVICE_INFO,
	  "setinfo",
	  &ngi_device_info_type,
	  NULL
	},
	{
	  NGM_IRLAP_COOKIE,
	  NGM_IRLAP_CMD_GET_DEVICE_INFO,
	  "getinfo",
	  NULL,
	  &ngi_device_info_type
	}
};

/* Define the netgraph node type */
static struct ng_type typestruct = {
	NG_VERSION,
	NG_IRLAP_NODE_TYPE,
	NULL,
	ngi_constructor,
	ngi_rcvmsg,
	ngi_shutdown,
	ngi_newhook,
	NULL,
	NULL,
	ngi_rcvdata,
	ngi_rcvdata,
	ngi_disconnect,
	ngi_cmdlist
};
NETGRAPH_INIT(irlap, &typestruct);

/******************************************************************
		     NETGRAPH NODE METHODS
******************************************************************/

/*
 * Initialize a new node
 */
static int
ngi_constructor(node_p *nodep)
{
	sc_p sc;
	int error;

	if ((error = ng_make_node_common(&typestruct, nodep)))
		 return (error);
	MALLOC(sc, sc_p, sizeof(*sc), M_NETGRAPH, M_WAITOK);
	if (sc == NULL)
		 return (ENOMEM);
	bzero(sc, sizeof(*sc));

	sc->conn_params.rate = NG_IRLAP_DEFAULT_RATE;
	sc->conn_params.maxturn = NG_IRLAP_DEFAULT_MAXTURN;
	sc->conn_params.datasize = NG_IRLAP_DEFAULT_DATASIZE;
	sc->conn_params.winsize = NG_IRLAP_DEFAULT_WINSIZE;
	sc->conn_params.nbof = NG_IRLAP_DEFAULT_NBOF;
	sc->conn_params.minturn = NG_IRLAP_DEFAULT_MINTURN;
	sc->conn_params.thresh = NG_IRLAP_DEFAULT_THRESH;

	sc->state = NG_IRLAP_STATE_NDM;
	sc->nrm_state = NG_IRLAP_NRM_RECV;

	sc->address = (u_int32_t) random();

	callout_handle_init(&sc->watchdog);

	(*nodep)->private = sc;
	sc->node = *nodep;
	return (0);
}

/*
 * Reserve a hook for a pending connection
 */
static int
ngi_newhook(node_p node, hook_p hook, const char *name)
{
        const sc_p sc = node->private;
        hook_p *hookp;

        if (!strcmp(name, NG_IRLAP_HOOK_UP))
                hookp = &sc->up;
        else if (!strcmp(name, NG_IRLAP_HOOK_DOWN))
                hookp = &sc->down;
        else
                return (EINVAL);
        if (*hookp)
                return (EISCONN);
        *hookp = hook;
        return (0);
}

/*
 * Receive incoming data
 */
static int
ngi_rcvdata(hook_p hook, struct mbuf *m, meta_p meta,
                struct mbuf **ret_m, meta_p *ret_meta)
{
        const sc_p sc = hook->node->private;

        if (hook == sc->up)
                /*return (ngi_rcv_up(sc, m, meta));*/
		return (EOPNOTSUPP);
        if (hook == sc->down)
                return (ngi_rcv_down(sc, m, meta));
        panic(__FUNCTION__);
}

/*
 * Receive incoming control message
 */
static int
ngi_rcvmsg(node_p node, struct ng_mesg *msg,
	const char *rtn, struct ng_mesg **rptr, hook_p lasthook)
{
	const sc_p sc = (sc_p)node->private;
	struct ng_mesg *resp = NULL;
	int error = 0;

	switch (msg->header.typecookie) {
	case NGM_IRLAP_COOKIE:
		switch (msg->header.cmd) {
		case NGM_IRLAP_CMD_SET_DEVICE_INFO:
		    {
			struct ng_irlap_device_info *const info =
				(struct ng_irlap_device_info *) msg->data;

			if (info->length > 23) {
				printf("ng_irlap: got info length %d\n",
				    info->length);
				ERROUT(EINVAL);
			}

			sc->device_info.length = info->length;
			bcopy(info->data, sc->device_info.data, info->length);

			break;
		    }
		case NGM_IRLAP_CMD_GET_DEVICE_INFO:
			NG_MKRESPONSE(resp, msg, sizeof(sc->device_info),
			    M_NOWAIT);
			if (!resp)
				ERROUT(ENOMEM);
			*((struct ng_irlap_device_info *) resp->data) =
			    sc->device_info;
			break;
		default:
			ERROUT(EINVAL);
		}
		break;
	default:
		ERROUT(EINVAL);
	}
	if (rptr)
		*rptr = resp;
	else if (resp)
		FREE(resp, M_NETGRAPH);

done:
	FREE(msg, M_NETGRAPH);
	return (error);
}

/*
 * Shutdown this node
 */
static int
ngi_shutdown(node_p node)
{
        const sc_p sc = node->private;

	ngi_stop_watchdog(sc);

        ng_cutlinks(node);
        ng_unname(node);

        bzero(sc, sizeof(*sc));
        FREE(sc, M_NETGRAPH);
        node->private = NULL;

        ng_unref(node);

        return (0);
}

/*
 * Lose a hook. When both hooks go away, we disappear.
 */
static int
ngi_disconnect(hook_p hook)
{
        const sc_p sc = hook->node->private;
        hook_p *hookp;

        if (hook == sc->down)
                hookp = &sc->down;
        else if (hook == sc->up)
                hookp = &sc->up;
        else
                panic(__FUNCTION__);
        if (!*hookp)
                panic(__FUNCTION__ "2");
        *hookp = NULL;
        if (hook->node->numhooks == 0)
                ng_rmnode(hook->node);
        return (0);
}

/******************************************************************
                    INTERNAL HELPER STUFF
******************************************************************/

/*
 * Receive frames from the lower layer.
 */
static int
ngi_rcv_down(const sc_p sc, struct mbuf *m, meta_p meta)
{
	int error;
	struct ng_irlap_frame *frame;

	error = 0;

	frame = mtod(m, struct ng_irlap_frame *);

	if ((frame->control & 0x1) == 0x0) {		/* I frame */
		if (sc->state != NG_IRLAP_STATE_NRM ||
		    (frame->connection & ~NG_IRLAP_MASK_CR) != sc->connection)
			ERROUT(ENOTCONN);

		return (ngi_process_info(sc, m, meta));
	} else if ((frame->control & 0x3) == 0x1) {	/* S frame */
		if (sc->state != NG_IRLAP_STATE_NRM ||
		    (frame->connection & ~NG_IRLAP_MASK_CR) != sc->connection)
			ERROUT(ENOTCONN);

		return (ngi_process_supervisory(sc, m, meta));
	} else {					/* U frame */
		return (ngi_process_unnumbered(sc, m, meta));
	}

done:
	NG_FREE_DATA(m, meta);
	return (error);
}

static int
ngi_process_unnumbered(const sc_p sc, struct mbuf *m, meta_p meta)
{
	struct ng_irlap_frame *frame;

	frame = mtod(m, struct ng_irlap_frame *);

	switch (frame->control & ~NG_IRLAP_MASK_PF) {
	case NG_IRLAP_COMMAND_XID:
		return(ngi_process_xid(sc, m, meta));
		break;
	case NG_IRLAP_COMMAND_SNRM:
		return(ngi_process_snrm(sc, m, meta));
		break;
	case NG_IRLAP_COMMAND_DISC:
		return(ngi_process_disc(sc, m, meta));
		break;
	default:
		NG_FREE_DATA(m, meta);
		return (EOPNOTSUPP);
		/* NOTREACHED */
	}
}

/*
 * Process U-XID frames.
 */
static int
ngi_process_xid(const sc_p sc, struct mbuf *m, meta_p meta)
{
	struct ifnet *const rcvif = m->m_pkthdr.rcvif;
	struct mbuf *outmbuf;
	struct ng_irlap_frame *frame, *outframe;
	struct ng_irlap_xid *xid, *outxid;
	u_int8_t format;
	int size, error;

	error = 0;

	frame = mtod(m, struct ng_irlap_frame *);
	format = *frame->data;
	if (format != 0x01) {
		NG_FREE_DATA(m, meta);
		return (EINVAL);
	}
	xid = (struct ng_irlap_xid *)(frame->data + 1);

	switch (sc->state) {
	case NG_IRLAP_STATE_NDM:
		switch (xid->flags) {
		case 0x03:
			sc->xid_slot = (u_int8_t)random() % 16;
			break;
		case 0x02:
			sc->xid_slot = (u_int8_t)random() % 8;
			break;
		case 0x01:
			sc->xid_slot = (u_int8_t)random() % 6;
			break;
		case 0x00:
		default:
			sc->xid_slot = (u_int8_t)random() % 1;
			break;
		}

		sc->state = NG_IRLAP_STATE_REPLY;

		if (sc->xid_slot != 0)
			break;
	case NG_IRLAP_STATE_REPLY:
		if (xid->slot <= sc->xid_slot) {
			size = sizeof(struct ng_irlap_frame) +
			    sizeof(struct ng_irlap_xid) +
			    sc->device_info.length + 1;

			MALLOC(outframe, struct ng_irlap_frame *, size,
			    M_NETGRAPH, M_NOWAIT);

			outframe->connection = NG_IRLAP_BROADCAST;
			outframe->control =
				NG_IRLAP_RESPONSE_XID | NG_IRLAP_MASK_PF;
			outframe->data[0] = 0x01;
			outxid = (struct ng_irlap_xid *)&outframe->data[1];

			outxid->destination = xid->source;
			outxid->source = sc->address;
			outxid->flags = 0;
			outxid->slot = 0;
			outxid->version = 0;
			bcopy(sc->device_info.data, outxid->data,
			    sc->device_info.length);

			if (!(outmbuf = m_devget((u_char *)outframe, size, 0,
			    rcvif, NULL))) {
				NG_FREE_DATA(m, meta);
				error = ENOBUFS;
			} else
				NG_SEND_DATA(error, sc->down, outmbuf, meta);

			sc->xid_slot = 0;
		}

		if (xid->slot == 0xff) {
			sc->state = NG_IRLAP_STATE_NDM;
		}

		break;
	default:
		NG_FREE_DATA(m, meta);
		return (EOPNOTSUPP);
		/* NOTREACHED */
	}

	NG_FREE_DATA(m, meta);
	return (error);
}

static int
ngi_process_snrm(const sc_p sc, struct mbuf *m, meta_p meta)
{
	struct ifnet *const rcvif = m->m_pkthdr.rcvif;
	struct mbuf *outmbuf;
	struct ng_irlap_frame *frame, *outframe;
	struct ng_irlap_snrm *snrm;
	struct ng_irlap_ua *ua;
	int size, error, i;

	error = 0;

	frame = mtod(m, struct ng_irlap_frame *);
	snrm = (struct ng_irlap_snrm *)frame->data;

	if (snrm->destination != sc->address)
		ERROUT(0);

	switch(sc->state) {
	case NG_IRLAP_STATE_NDM:
		sc->connection = snrm->connection & ~NG_IRLAP_MASK_CR;

		size = sizeof(struct ng_irlap_frame) +
		    sizeof(struct ng_irlap_ua);
		MALLOC(outframe, struct ng_irlap_frame *, size,
		    M_NETGRAPH, M_NOWAIT);

		bzero(outframe, size);
		outframe->connection = snrm->connection;
		outframe->control = NG_IRLAP_RESPONSE_UA | NG_IRLAP_MASK_PF;

		ua = (struct ng_irlap_ua *)outframe->data;
		ua->source = sc->address;
		ua->destination = snrm->source;

		/* XXX: Negotiation needs to be done properly. */
		i = 0;
		ua->data[i++] = NG_IRLAP_PARAM_RATE;
		ua->data[i++] = sizeof(sc->conn_params.rate);
		ua->data[i++] = sc->conn_params.rate >> 16;
		ua->data[i++] = sc->conn_params.rate & 0x00ff;

		ua->data[i++] = NG_IRLAP_PARAM_MAXTURN;
		ua->data[i++] = sizeof(sc->conn_params.maxturn);
		ua->data[i++] = sc->conn_params.maxturn;

		ua->data[i++] = NG_IRLAP_PARAM_DATASIZE;
		ua->data[i++] = sizeof(sc->conn_params.datasize);
		ua->data[i++] = sc->conn_params.datasize;

		ua->data[i++] = NG_IRLAP_PARAM_WINSIZE;
		ua->data[i++] = sizeof(sc->conn_params.winsize);
		ua->data[i++] = sc->conn_params.winsize;

		ua->data[i++] = NG_IRLAP_PARAM_NBOF;
		ua->data[i++] = sizeof(sc->conn_params.nbof);
		ua->data[i++] = sc->conn_params.nbof;

		ua->data[i++] = NG_IRLAP_PARAM_MINTURN;
		ua->data[i++] = sizeof(sc->conn_params.minturn);
		ua->data[i++] = sc->conn_params.minturn;

		ua->data[i++] = NG_IRLAP_PARAM_THRESH;
		ua->data[i++] = sizeof(sc->conn_params.thresh);
		ua->data[i++] = sc->conn_params.thresh;

		if (!(outmbuf = m_devget((u_char *)outframe, size, 0,
		    rcvif, NULL))) {
			error = ENOBUFS;
		} else {
			NG_SEND_DATA(error, sc->down, outmbuf, meta);
			sc->state = NG_IRLAP_STATE_NRM;
			sc->primary = 0;
		}

		break;
	}

done:
	NG_FREE_DATA(m, meta);
	return(error);
}

static int
ngi_process_disc(const sc_p sc, struct mbuf *m, meta_p meta)
{
	struct ifnet *const rcvif = m->m_pkthdr.rcvif;
	struct mbuf *outmbuf;
	struct ng_irlap_frame *outframe;
	int error;

	error = 0;

	switch(sc->state) {
	case NG_IRLAP_STATE_NRM:
		MALLOC(outframe, struct ng_irlap_frame *,
		    sizeof(struct ng_irlap_frame), M_NETGRAPH, M_NOWAIT);

		bzero(outframe, sizeof(struct ng_irlap_frame));
		outframe->connection = sc->connection;
		outframe->control = NG_IRLAP_RESPONSE_UA | NG_IRLAP_MASK_PF;

		if (!(outmbuf = m_devget((u_char *)outframe,
		    sizeof(struct ng_irlap_frame), 0, rcvif, NULL))) {
			ERROUT(ENOBUFS);
		} else {
			NG_SEND_DATA(error, sc->down, outmbuf, meta);
			sc->connection = 0;
		}

		break;
	}

done:
	NG_FREE_DATA(m, meta);
	return (error);
}

/* Process an information frame */
static int
ngi_process_info(const sc_p sc, struct mbuf *m, meta_p meta)
{
	int error;
	struct ng_irlap_frame *frame;
	int nr, ns, pf;

	error = 0;

	frame = mtod(m, struct ng_irlap_frame *);

	if ((frame->connection & NG_IRLAP_MASK_CR) == sc->primary) {
		/* RECV state 25 */
		ERROUT(EINVAL);
	}

	pf = NG_IRLAP_GET_PF(frame->control);
	nr = NG_IRLAP_GET_NR(frame->control);
	ns = NG_IRLAP_GET_NS(frame->control);

	if (ns == sc->in_seq) {
		if (pf) {
			/* RECV state 1, 11 */
		} else {
			/* RECV state 2, 3 */
		}
	} else {
		if (pf) {
			/* RECV state 9 */
		} else {
			/* RECV state 10 */
		}
	}

done:
	NG_FREE_DATA(m, meta);
	return (error);
}

/* Process a supervisory frame */
static int
ngi_process_supervisory(const sc_p sc, struct mbuf *m, meta_p meta)
{
	int error;
	struct ng_irlap_frame *frame;
	int nr, pf, scmd;

	error = 0;

	frame = mtod(m, struct ng_irlap_frame *);

	if ((frame->connection & NG_IRLAP_MASK_CR) == sc->primary) {
		/* RECV state 25 */
		ERROUT(EINVAL);
	}

	pf = NG_IRLAP_GET_PF(frame->control);
	nr = NG_IRLAP_GET_NR(frame->control);
	scmd = NG_IRLAP_GET_SCMD(frame->control);

	switch (scmd) {
	case NG_IRLAP_S_RR:
		if (/* invalid/unexpected Nr */ 0) {
			/* RECV state 12 */

			sc->remote_busy = 0;
			/* sc->in_seq = nr + 1; */

			/* Resend frames? */

			ngi_start_watchdog(sc);
		} else if (sc->remote_busy || 1) {	/* || no-pending-data-reqs */
			/* RECV state 17 */

			sc->remote_busy = 0;
			/* update Nr received */
			/* turnaround delay? */
			error = ngi_send_supervisory(sc, NG_IRLAP_S_RR, 0, nr,
			    1);
			ngi_start_watchdog(sc);
		} else {
			/* RECV state 16 */

			sc->remote_busy = 0;
			/* update Nr received */
			ngi_stop_watchdog(sc);
			sc->nrm_state = NG_IRLAP_NRM_XMIT;
		}

		break;
	case NG_IRLAP_S_REJ:
		/* RECV state 13, 18 */

		break;
	case NG_IRLAP_S_RNR:
		/* RECV state 19 */

		break;
	case NG_IRLAP_S_SREJ:
		break;
	}

done:
	NG_FREE_DATA(m, meta);
	return (error);
}

static int
ngi_send_supervisory(const sc_p sc, u_int8_t command, u_int8_t cr, u_int8_t vr,
    u_int8_t pf)
{
	int error;
	struct ng_irlap_frame *frame;
	struct mbuf *m;
	meta_p meta;

	error = 0;
	meta = NULL;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		ERROUT(ENOBUFS);

	frame = (struct ng_irlap_frame *)m->m_data;
	frame->connection = sc->connection | cr;
	frame->control = (command << 2) | (pf << 4) | (vr << 5);

	m->m_len = m->m_pkthdr.len = sizeof(struct ng_irlap_frame);
	m->m_pkthdr.rcvif = NULL;

	NG_SEND_DATA(error, sc->down, m, meta);

done:
	return (error);
}

static void
ngi_start_watchdog(const sc_p sc)
{
	untimeout(ngi_watchdog_expired, (void *)sc, sc->watchdog);

	sc->watchdog = timeout(ngi_watchdog_expired, (void *)sc,
	    NG_IRLAP_TIMEOUT_WATCHDOG);
}

static void
ngi_stop_watchdog(const sc_p sc)
{
	untimeout(ngi_watchdog_expired, (void *)sc, sc->watchdog);
	callout_handle_init(&sc->watchdog);
}

static void
ngi_watchdog_expired(void *arg)
{
	sc_p sc = (sc_p)arg;

	sc->watchdog_expires++;
	printf("ng_irlap: watchdog timer expired, count is %d\n",
	    sc->watchdog_expires);

	if (sc->watchdog_expires < 20)
		ngi_start_watchdog(sc);
}
